from __future__ import print_function, division, absolute_import

from distributed.protocol.utils import merge_frames, pack_frames, unpack_frames
from distributed.utils import ensure_bytes


def test_merge_frames():
    result = merge_frames({'lengths': [3, 4]}, [b'12', b'34', b'567'])
    expected = [b'123', b'4567']

    assert list(map(ensure_bytes, result)) == expected

    b = b'123'
    assert merge_frames({'lengths': [3]}, [b])[0] is b

    L = [b'123', b'456']
    assert merge_frames({'lengths': [3, 3]}, L) is L


def test_pack_frames():
    frames = [b'123', b'asdf']
    b = pack_frames(frames)
    assert isinstance(b, bytes)
    frames2 = unpack_frames(b)

    assert frames == frames2



from __future__ import print_function, division, absolute_import

import random
import time

from concurrent.futures import (
    CancelledError, TimeoutError, Future, wait, as_completed,
    FIRST_COMPLETED, FIRST_EXCEPTION)

import pytest
from toolz import take

from distributed.client import Client
from distributed.utils_test import (slowinc, slowadd, slowdec,
                                    inc, throws, cluster)
from distributed.utils_test import loop # flake8: noqa


def number_of_processing_tasks(client):
    return sum(len(v) for k, v in client.processing().items())


def test_submit(loop):
    with cluster() as (s, [a, b]):
        with Client(s['address'], loop=loop) as c:
            with c.get_executor() as e:
                f1 = e.submit(slowadd, 1, 2)
                assert isinstance(f1, Future)
                f2 = e.submit(slowadd, 3, y=4)
                f3 = e.submit(throws, "foo")
                f4 = e.submit(slowadd, x=5, y=6)
                assert f1.result() == 3
                assert f2.result() == 7
                with pytest.raises(RuntimeError):
                    f3.result()
                assert f4.result() == 11


def test_as_completed(loop):
    with cluster() as (s, [a, b]):
        with Client(s['address'], loop=loop) as c:
            with c.get_executor() as e:
                N = 10
                fs = [e.submit(slowinc, i, delay=0.02) for i in range(N)]
                expected = set(range(1, N + 1))

                for f in as_completed(fs):
                    res = f.result()
                    assert res in expected
                    expected.remove(res)

                assert not expected


def test_wait(loop):
    with cluster() as (s, [a, b]):
        with Client(s['address'], loop=loop) as c:
            with c.get_executor(pure=False) as e:
                N = 10
                fs = [e.submit(slowinc, i, delay=0.05) for i in range(N)]
                res = wait(fs, timeout=0.01)
                assert len(res.not_done) > 0
                res = wait(fs)
                assert len(res.not_done) == 0
                assert res.done == set(fs)

                fs = [e.submit(slowinc, i, delay=0.05) for i in range(N)]
                res = wait(fs, return_when=FIRST_COMPLETED)
                assert len(res.not_done) > 0
                assert len(res.done) >= 1
                res = wait(fs)
                assert len(res.not_done) == 0
                assert res.done == set(fs)

                fs = [e.submit(slowinc, i, delay=0.05) for i in range(N)]
                fs += [e.submit(throws, None)]
                fs += [e.submit(slowdec, i, delay=0.05) for i in range(N)]
                res = wait(fs, return_when=FIRST_EXCEPTION)
                assert len(res.not_done) > 0
                assert len(res.done) >= N - 2  # likely, unless tasks get reordered

                errors = []
                for fs in res.done:
                    try:
                        fs.result()
                    except RuntimeError as e:
                        errors.append(e)

                assert len(errors) == 1
                assert "hello" in str(errors[0])


def test_cancellation(loop):
    with cluster() as (s, [a, b]):
        with Client(s['address'], loop=loop) as c:
            with c.get_executor(pure=False) as e:
                fut = e.submit(time.sleep, 2.0)
                start = time.time()
                while number_of_processing_tasks(c) == 0:
                    assert time.time() < start + 1
                    time.sleep(0.01)
                assert not fut.done()

                fut.cancel()
                assert fut.cancelled()
                start = time.time()
                while number_of_processing_tasks(c) != 0:
                    assert time.time() < start + 1
                    time.sleep(0.01)

                with pytest.raises(CancelledError):
                    fut.result()

            # With wait()
            with c.get_executor(pure=False) as e:
                N = 10
                fs = [e.submit(slowinc, i, delay=0.02) for i in range(N)]
                fs[3].cancel()
                res = wait(fs, return_when=FIRST_COMPLETED)
                assert len(res.not_done) > 0
                assert len(res.done) >= 1

                assert fs[3] in res.done
                assert fs[3].cancelled()

            # With as_completed()
            with c.get_executor(pure=False) as e:
                N = 10
                fs = [e.submit(slowinc, i, delay=0.02) for i in range(N)]
                fs[3].cancel()
                fs[8].cancel()

                n_cancelled = sum(f.cancelled() for f in as_completed(fs))
                assert n_cancelled == 2


def test_map(loop):
    with cluster() as (s, [a, b]):
        with Client(s['address'], loop=loop) as c:
            with c.get_executor() as e:
                N = 10
                it = e.map(inc, range(N))
                expected = set(range(1, N + 1))
                for x in it:
                    expected.remove(x)
                assert not expected

            with c.get_executor(pure=False) as e:
                N = 10
                it = e.map(slowinc, range(N), [0.1] * N, timeout=0.4)
                results = []
                with pytest.raises(TimeoutError):
                    for x in it:
                        results.append(x)
                assert 2 <= len(results) < 7

            with c.get_executor(pure=False) as e:
                N = 10
                # Not consuming the iterator will cancel remaining tasks
                it = e.map(slowinc, range(N), [0.1] * N)
                for x in take(2, it):
                    pass
                # Some tasks still processing
                assert number_of_processing_tasks(c) > 0
                # Garbage collect the iterator => remaining tasks are cancelled
                del it
                assert number_of_processing_tasks(c) == 0


def get_random():
    return random.random()


def test_pure(loop):
    with cluster() as (s, [a, b]):
        with Client(s['address'], loop=loop) as c:
            N = 10
            with c.get_executor() as e:
                fs = [e.submit(get_random) for i in range(N)]
                res = [fut.result() for fut in as_completed(fs)]
                assert len(set(res)) < len(res)
            with c.get_executor(pure=False) as e:
                fs = [e.submit(get_random) for i in range(N)]
                res = [fut.result() for fut in as_completed(fs)]
                assert len(set(res)) == len(res)


def test_workers(loop):
    with cluster() as (s, [a, b]):
        with Client(s['address'], loop=loop) as c:
            N = 10
            with c.get_executor(workers=[b['address']]) as e:
                fs = [e.submit(slowinc, i) for i in range(N)]
                wait(fs)
                has_what = c.has_what()
                assert not has_what.get(a['address'])
                assert len(has_what[b['address']]) == N


def test_unsupported_arguments(loop):
    with cluster() as (s, [a, b]):
        with Client(s['address'], loop=loop) as c:
            with pytest.raises(TypeError) as excinfo:
                c.get_executor(workers=[b['address']], foo=1, bar=2)
            assert ("unsupported arguments to ClientExecutor: ['bar', 'foo']"
                    in str(excinfo.value))


def test_shutdown(loop):
    with cluster(active_rpc_timeout=10) as (s, [a, b]):
        with Client(s['address'], loop=loop) as c:
            # shutdown(wait=True) waits for pending tasks to finish
            e = c.get_executor()
            fut = e.submit(time.sleep, 1.0)
            t1 = time.time()
            e.shutdown()
            dt = time.time() - t1
            assert 0.5 <= dt <= 2.0
            time.sleep(0.1)  # wait for future outcome to propagate
            assert fut.done()
            fut.result()  # doesn't raise

            with pytest.raises(RuntimeError):
                e.submit(time.sleep, 1.0)

            # shutdown(wait=False) cancels pending tasks
            e = c.get_executor()
            fut = e.submit(time.sleep, 2.0)
            t1 = time.time()
            e.shutdown(wait=False)
            dt = time.time() - t1
            assert dt < 0.5
            time.sleep(0.1)  # wait for future outcome to propagate
            assert fut.cancelled()

            with pytest.raises(RuntimeError):
                e.submit(time.sleep, 1.0)



from __future__ import print_function, division, absolute_import

from contextlib import contextmanager
import sys

try:
    import ssl
except ImportError:
    ssl = None

import pytest
from tornado import gen

from distributed.comm import connect, listen
from distributed.security import Security
from distributed.utils_test import new_config, get_cert, gen_test


ca_file = get_cert('tls-ca-cert.pem')

cert1 = get_cert('tls-cert.pem')
key1 = get_cert('tls-key.pem')
keycert1 = get_cert('tls-key-cert.pem')

# Note this cipher uses RSA auth as this matches our test certs
FORCED_CIPHER = 'ECDHE-RSA-AES128-GCM-SHA256'


def test_defaults():
    with new_config({}):
        sec = Security()
    assert sec.require_encryption in (None, False)
    assert sec.tls_ca_file is None
    assert sec.tls_ciphers is None
    assert sec.tls_client_key is None
    assert sec.tls_client_cert is None
    assert sec.tls_scheduler_key is None
    assert sec.tls_scheduler_cert is None
    assert sec.tls_worker_key is None
    assert sec.tls_worker_cert is None


def test_attribute_error():
    sec = Security()
    assert hasattr(sec, 'tls_ca_file')
    with pytest.raises(AttributeError):
        sec.tls_foobar
    with pytest.raises(AttributeError):
        sec.tls_foobar = ""


def test_from_config():
    c = {
        'tls': {
            'ca-file': 'ca.pem',
            'scheduler': {
                'key': 'skey.pem',
                'cert': 'scert.pem',
            },
            'worker': {
                'cert': 'wcert.pem',
            },
            'ciphers': FORCED_CIPHER,
        },
        'require-encryption': True,
    }
    with new_config(c):
        sec = Security()
    assert sec.require_encryption is True
    assert sec.tls_ca_file == 'ca.pem'
    assert sec.tls_ciphers == FORCED_CIPHER
    assert sec.tls_client_key is None
    assert sec.tls_client_cert is None
    assert sec.tls_scheduler_key == 'skey.pem'
    assert sec.tls_scheduler_cert == 'scert.pem'
    assert sec.tls_worker_key is None
    assert sec.tls_worker_cert == 'wcert.pem'


def test_kwargs():
    c = {
        'tls': {
            'ca-file': 'ca.pem',
            'scheduler': {
                'key': 'skey.pem',
                'cert': 'scert.pem',
            },
        },
    }
    with new_config(c):
        sec = Security(tls_scheduler_cert='newcert.pem',
                       require_encryption=True,
                       tls_ca_file=None)
    assert sec.require_encryption is True
    # None value didn't override default
    assert sec.tls_ca_file == 'ca.pem'
    assert sec.tls_ciphers is None
    assert sec.tls_client_key is None
    assert sec.tls_client_cert is None
    assert sec.tls_scheduler_key == 'skey.pem'
    assert sec.tls_scheduler_cert == 'newcert.pem'
    assert sec.tls_worker_key is None
    assert sec.tls_worker_cert is None


def test_repr():
    with new_config({}):
        sec = Security(tls_ca_file='ca.pem', tls_scheduler_cert='scert.pem')
        assert repr(sec) == "Security(tls_ca_file='ca.pem', tls_scheduler_cert='scert.pem')"


def test_tls_config_for_role():
    c = {
        'tls': {
            'ca-file': 'ca.pem',
            'scheduler': {
                'key': 'skey.pem',
                'cert': 'scert.pem',
            },
            'worker': {
                'cert': 'wcert.pem',
            },
            'ciphers': FORCED_CIPHER,
        },
    }
    with new_config(c):
        sec = Security()
    t = sec.get_tls_config_for_role('scheduler')
    assert t == {
        'ca_file': 'ca.pem',
        'key': 'skey.pem',
        'cert': 'scert.pem',
        'ciphers': FORCED_CIPHER,
    }
    t = sec.get_tls_config_for_role('worker')
    assert t == {
        'ca_file': 'ca.pem',
        'key': None,
        'cert': 'wcert.pem',
        'ciphers': FORCED_CIPHER,
    }
    t = sec.get_tls_config_for_role('client')
    assert t == {
        'ca_file': 'ca.pem',
        'key': None,
        'cert': None,
        'ciphers': FORCED_CIPHER,
    }
    with pytest.raises(ValueError):
        sec.get_tls_config_for_role('supervisor')


def test_connection_args():
    def basic_checks(ctx):
        assert ctx.verify_mode == ssl.CERT_REQUIRED
        assert ctx.check_hostname is False

    def many_ciphers(ctx):
        if sys.version_info >= (3, 6):
            assert len(ctx.get_ciphers()) > 2  # Most likely

    c = {
        'tls': {
            'ca-file': ca_file,
            'scheduler': {
                'key': key1,
                'cert': cert1,
            },
            'worker': {
                'cert': keycert1,
            },
        },
    }
    with new_config(c):
        sec = Security()

    d = sec.get_connection_args('scheduler')
    assert not d['require_encryption']
    ctx = d['ssl_context']
    basic_checks(ctx)
    many_ciphers(ctx)

    d = sec.get_connection_args('worker')
    ctx = d['ssl_context']
    basic_checks(ctx)
    many_ciphers(ctx)

    # No cert defined => no TLS
    d = sec.get_connection_args('client')
    assert d.get('ssl_context') is None

    # With more settings
    c['tls']['ciphers'] = FORCED_CIPHER
    c['require-encryption'] = True

    with new_config(c):
        sec = Security()

    d = sec.get_listen_args('scheduler')
    assert d['require_encryption']
    ctx = d['ssl_context']
    basic_checks(ctx)
    if sys.version_info >= (3, 6):
        assert len(ctx.get_ciphers()) == 1


def test_listen_args():
    def basic_checks(ctx):
        assert ctx.verify_mode == ssl.CERT_REQUIRED
        assert ctx.check_hostname is False

    def many_ciphers(ctx):
        if sys.version_info >= (3, 6):
            assert len(ctx.get_ciphers()) > 2  # Most likely

    c = {
        'tls': {
            'ca-file': ca_file,
            'scheduler': {
                'key': key1,
                'cert': cert1,
            },
            'worker': {
                'cert': keycert1,
            },
        },
    }
    with new_config(c):
        sec = Security()

    d = sec.get_listen_args('scheduler')
    assert not d['require_encryption']
    ctx = d['ssl_context']
    basic_checks(ctx)
    many_ciphers(ctx)

    d = sec.get_listen_args('worker')
    ctx = d['ssl_context']
    basic_checks(ctx)
    many_ciphers(ctx)

    # No cert defined => no TLS
    d = sec.get_listen_args('client')
    assert d.get('ssl_context') is None

    # With more settings
    c['tls']['ciphers'] = FORCED_CIPHER
    c['require-encryption'] = True

    with new_config(c):
        sec = Security()

    d = sec.get_listen_args('scheduler')
    assert d['require_encryption']
    ctx = d['ssl_context']
    basic_checks(ctx)
    if sys.version_info >= (3, 6):
        assert len(ctx.get_ciphers()) == 1


@gen_test()
def test_tls_listen_connect():
    """
    Functional test for TLS connection args.
    """
    @gen.coroutine
    def handle_comm(comm):
        peer_addr = comm.peer_address
        assert peer_addr.startswith('tls://')
        yield comm.write('hello')
        yield comm.close()

    c = {
        'tls': {
            'ca-file': ca_file,
            'scheduler': {
                'key': key1,
                'cert': cert1,
            },
            'worker': {
                'cert': keycert1,
            },
        },
    }
    with new_config(c):
        sec = Security()

    c['tls']['ciphers'] = FORCED_CIPHER
    with new_config(c):
        forced_cipher_sec = Security()

    with listen('tls://', handle_comm,
                connection_args=sec.get_listen_args('scheduler')) as listener:
        comm = yield connect(listener.contact_address,
                             connection_args=sec.get_connection_args('worker'))
        msg = yield comm.read()
        assert msg == 'hello'
        comm.abort()

        # No SSL context for client
        with pytest.raises(TypeError):
            yield connect(listener.contact_address,
                          connection_args=sec.get_connection_args('client'))

        # Check forced cipher
        comm = yield connect(listener.contact_address,
                             connection_args=forced_cipher_sec.get_connection_args('worker'))
        cipher, _, _, = comm.extra_info['cipher']
        assert cipher == FORCED_CIPHER
        comm.abort()


@gen_test()
def test_require_encryption():
    """
    Functional test for "require_encryption" setting.
    """
    @gen.coroutine
    def handle_comm(comm):
        comm.abort()

    c = {
        'tls': {
            'ca-file': ca_file,
            'scheduler': {
                'key': key1,
                'cert': cert1,
            },
            'worker': {
                'cert': keycert1,
            },
        },
    }
    with new_config(c):
        sec = Security()
    c['require-encryption'] = True
    with new_config(c):
        sec2 = Security()

    for listen_addr in ['inproc://', 'tls://']:
        with listen(listen_addr, handle_comm,
                    connection_args=sec.get_listen_args('scheduler')) as listener:
            comm = yield connect(listener.contact_address,
                                 connection_args=sec2.get_connection_args('worker'))
            comm.abort()

        with listen(listen_addr, handle_comm,
                    connection_args=sec2.get_listen_args('scheduler')) as listener:
            comm = yield connect(listener.contact_address,
                                 connection_args=sec2.get_connection_args('worker'))
            comm.abort()

    @contextmanager
    def check_encryption_error():
        with pytest.raises(RuntimeError) as excinfo:
            yield
        assert "encryption required" in str(excinfo.value)

    for listen_addr in ['tcp://']:
        with listen(listen_addr, handle_comm,
                    connection_args=sec.get_listen_args('scheduler')) as listener:
            comm = yield connect(listener.contact_address,
                                 connection_args=sec.get_connection_args('worker'))
            comm.abort()

            with pytest.raises(RuntimeError):
                yield connect(listener.contact_address,
                              connection_args=sec2.get_connection_args('worker'))

        with pytest.raises(RuntimeError):
            listen(listen_addr, handle_comm,
                   connection_args=sec2.get_listen_args('scheduler'))



from collections import Iterator
from operator import add
import random
from time import sleep

import pytest
from tornado import gen

from distributed import Client
from distributed.client import _as_completed, as_completed, _first_completed
from distributed.compatibility import Empty
from distributed.utils_test import cluster, gen_cluster, inc
from distributed.utils_test import loop # flake8: noqa
from distributed.compatibility import Queue


@gen_cluster(client=True)
def test__as_completed(c, s, a, b):
    x = c.submit(inc, 1)
    y = c.submit(inc, 1)
    z = c.submit(inc, 2)

    queue = Queue()
    yield _as_completed([x, y, z], queue)

    assert queue.qsize() == 3
    assert {queue.get(), queue.get(), queue.get()} == {x, y, z}

    result = yield _first_completed([x, y, z])
    assert result in [x, y, z]


def test_as_completed(loop):
    with cluster() as (s, [a, b]):
        with Client(s['address'], loop=loop) as c:
            x = c.submit(inc, 1)
            y = c.submit(inc, 2)
            z = c.submit(inc, 1)

            seq = as_completed([x, y, z])
            assert isinstance(seq, Iterator)
            assert set(seq) == {x, y, z}

            assert list(as_completed([])) == []


def test_as_completed_with_non_futures(loop):
    with cluster() as (s, [a, b]):
        with Client(s['address'], loop=loop):
            with pytest.raises(TypeError):
                list(as_completed([1, 2, 3]))


def test_as_completed_add(loop):
    with cluster() as (s, [a, b]):
        with Client(s['address'], loop=loop) as c:
            total = 0
            expected = sum(map(inc, range(10)))
            futures = c.map(inc, range(10))
            ac = as_completed(futures)
            for future in ac:
                result = future.result()
                total += result
                if random.random() < 0.5:
                    future = c.submit(add, future, 10)
                    ac.add(future)
                    expected += result + 10
            assert total == expected


def test_as_completed_update(loop):
    with cluster() as (s, [a, b]):
        with Client(s['address'], loop=loop) as c:
            total = 0
            todo = list(range(10))
            expected = sum(map(inc, todo))
            ac = as_completed([])
            while todo or not ac.is_empty():
                if todo:
                    work, todo = todo[:4], todo[4:]
                    ac.update(c.map(inc, work))
                batch = ac.next_batch(block=True)
                total += sum(r.result() for r in batch)
            assert total == expected


def test_as_completed_repeats(loop):
    with cluster() as (s, [a, b]):
        with Client(s['address'], loop=loop) as c:
            ac = as_completed()
            x = c.submit(inc, 1)
            ac.add(x)
            ac.add(x)

            assert next(ac) is x
            assert next(ac) is x

            with pytest.raises(StopIteration):
                next(ac)

            ac.add(x)
            assert next(ac) is x


def test_as_completed_is_empty(loop):
    with cluster() as (s, [a, b]):
        with Client(s['address'], loop=loop) as c:
            ac = as_completed()
            assert ac.is_empty()
            x = c.submit(inc, 1)
            ac.add(x)
            assert not ac.is_empty()
            assert next(ac) is x
            assert ac.is_empty()


def test_as_completed_cancel(loop):
    with cluster() as (s, [a, b]):
        with Client(s['address'], loop=loop) as c:
            x = c.submit(sleep, 1)
            y = c.submit(inc, 1)

            ac = as_completed([x, y])
            x.cancel()

            assert next(ac) is y

            with pytest.raises(Empty):
                ac.queue.get(timeout=0.1)

            assert list(as_completed([x, y, x])) == [y]


def test_as_completed_cancel_last(loop):
    with cluster() as (s, [a, b]):
        with Client(s['address'], loop=loop) as c:
            w = c.submit(sleep, 0.3)
            x = c.submit(inc, 1)
            y = c.submit(sleep, 0.3)

            @gen.coroutine
            def _():
                yield gen.sleep(0.1)
                yield w.cancel(asynchronous=True)
                yield y.cancel(asynchronous=True)

            loop.add_callback(_)

            ac = as_completed([x, y])
            result = list(ac)

            assert result == [x]
